/* 
* 
* Module Name : upnp-pcp iwf
 
* Version : 01
 
*
 
* Software Name : upnp-pcp interworking function according to pcp base draft -12
 
* Version : 01
 
*
 
 
 
* Copyright: 10/06/2011  France Telecom
 
* This software is distributed under the Apache 2.0 license,
 
* the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 
 
 
*
 
*--------------------------------------------------------
 
* File Name   : pcp.c
 
*
 
* Created     : 03/2011
 
* Author(s)   : Xiaohong Deng, Tao Zheng, Zhengyu Lei
 
*
 
* Description :
 
* pcp_log : record pcp error information
* add_mapping_local : add mapping (five tuples, lifetime and descriptions)  to memory
* get_cpe_mapping : get all mappings from CGN
* del_mapping_local : delete mapping from memory
* get_mapping_local : get a specific mapping from memory
* get_mapping : get a specific mapping from CGN
* add_rdrlist_CGN : add local mappings to CGN for Server rebuilding
* add_mapping_CGN : send a specific pcp add-mapping request to CGN
* get_mapping_by_index : get a specific mapping by index
* delete_mapping_CGN : send a specific pcp delete-mapping request to CGN
* delete_mappings : send request to delete mappings
* del_all_mappings_CGN : send request to delete all mappings from this CPE in CGN
* refresh_mappings : refresh mappings 
*--------------------------------------------------------
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "pcp.h"
#include "upnpglobalvars.h"

#define GET_OPCODE 0x62//ZTE set this 0x60, HUAWEI set this 0x62
#define DESCRIPTION_OPTION 0x40
#define NEXT_OPTION 0x83



//static unsigned short cnt;
static uint32_t lasttm;
struct epoch_timer epoch_s;

int pcp_log(const char *string){
	int fd;
	if((fd = open("PCP_error.log", O_RDWR|O_CREAT|O_APPEND, 0666)) == -1){
		printf("open:%s  with errno:%d\n", strerror(errno), errno);
		return -1;
	}


	if(write(fd, string, strlen(string)) != strlen(string)){
		printf("write:%s  with errno:%d\n", strerror(errno), errno);

		return -1;
	}
	
	close(fd);
	return 1;
}


/*disply the PCP packet content*/
void print_packet_content(unsigned char *pBuf, int bufLen)
{
    int count = 0;    

    printf("PCP packet content:\n");
    
    if ((NULL == pBuf) || (0 == bufLen))
    {
        printf("null\n");
        return;
    }

    for (count = 0; count < bufLen; count++)
    {
        printf("%2x ", pBuf[count]);
        if ((count+1) % 8 == 0 )
        {
            printf("\n");
        }
    }
	printf("\n");

    return;
}


static struct rdr *
add_mapping_local(unsigned short ex_port, const char *in_ip, const char *ex_ip,
				  unsigned short in_port, int protocol,
				  uint32_t lifetime, const char *description)

{
	struct rdr * p;
	size_t l = 1;


	#ifdef DEBUG_INFO
	printf("*****add_redirect_internal:\n");
	syslog(LOG_WARNING, "*****add_redirect_internal:");
	#endif

	if (description)
		l += strlen(description);
	p = malloc(sizeof(struct rdr) + l);
	if (p) {
		memset(p, 0, sizeof(struct rdr));
		p->next = rdr_list;
		strncpy(p->iaddr, in_ip, INET_ADDRSTRLEN);
		strncpy(p->eaddr, ex_ip, INET_ADDRSTRLEN);
		p->proto = (short)protocol;
		p->iport = in_port;
		p->eport = ex_port;
		p->lifetime = lifetime;
		if (lifetime)
			p->expire = (uint32_t)
				(time(NULL) - startup_time) + lifetime;
		if (description)
			memcpy(p->desc, description, l);
		rdr_list = p;
	}

	return p;
}
int get_cpe_mapping(void)
{
	
	unsigned char sbuf[256]/* 28bytes common-response-header + 40bytes GET opcode + 44bytesone next-option + 36bytes description-option*/, rbuf[1024];

	unsigned short rcv_descr_len = 0, first_next_len =0, des_len_in4 = 0;
	int recv_len, error;

	int final_next_len =0;
		
	memset(sbuf, 0, sizeof(sbuf));
	memset(rbuf, 0, sizeof(rbuf));

	int lifetime = 0;
	unsigned short s_inport = 0, s_export = 0;
	int proto= 0;
	char descr_buffer[128];
	char next_option[1024];


	memset(next_option, 0, sizeof(next_option));
	
	memset(descr_buffer, 0, sizeof(descr_buffer));



	sbuf[0] = 0x01;//version value (In 5th-version it's value is OX80)
	sbuf[1] = GET_OPCODE;//GET opcode 
	struct sockaddr_in6 pcp_cli_ip;
	memset(&pcp_cli_ip, 0, sizeof(pcp_cli_ip));
	
	if (inet_pton(AF_INET6, xnatpmp_local, &pcp_cli_ip.sin6_addr) <= 0) {
		#ifdef DEBUG_INFO
		printf("Failed to convert '%s' to local IPv6 address\n", xnatpmp_local);
		syslog(LOG_ERR, "Failed to convert '%s' to local IPv6 address", xnatpmp_local);
		#endif
		return -2;
	}

	memcpy(sbuf +12, &pcp_cli_ip.sin6_addr, 16);

	sbuf[29] =1;//in-AF
	sbuf[30] =1;//ex-AF

	sbuf[68] = NEXT_OPTION;

	first_next_len = 48;

	first_next_len = htons(first_next_len);

	memcpy(sbuf+70, &first_next_len, sizeof(first_next_len));


	sbuf[73]=1;
	sbuf[74]=1;

	sbuf[116]=DESCRIPTION_OPTION;
	final_next_len = 52;
	

	char *now = NULL;
	char c_exip_4[INET_ADDRSTRLEN], c_inip_4[INET_ADDRSTRLEN];
	char c_exip_6[INET6_ADDRSTRLEN], c_inip_6[INET6_ADDRSTRLEN];
	char in_af = 0 , ex_af = 0;
	char flag ='\0', descr_flag = '\0';
	int flag_len = 0, d_len =0;
	int next_len =0;
	#ifdef DEBUG_INFO
	int i;
	#endif

	

	re_send:

	#ifdef DEBUG_INFO
	printf("&&&&&GET send :\n");
	print_packet_content((unsigned char *)sbuf, 68+final_next_len);
	#endif

	if (send(sxfd6, sbuf, 68+final_next_len, 0) < 0){//send to CGN the sxfd6
			syslog(LOG_WARNING, "get_CPE_mappings(send): %m");
			return -1;
	}

	memset(rbuf, 0, sizeof(rbuf));
	if((recv_len = recv(sxfd6, rbuf, sizeof(rbuf), 0)) < 0){
		syslog(LOG_WARNING, "get_all_mappings(recv): %m");
		return -1;
	}

	flag_len = recv_len -68;

	#ifdef DEBUG_INFO
	printf("\n\t****GET recv from CGN:\n");
	print_packet_content(rbuf, recv_len);
	#endif
	
	error = rbuf[3];
	printf("$$$$$$\n\terror code : %d\n", error);
	if(error != 0){
		memcpy(&epoch_s.lastepoch, rbuf + 8, 4);
		epoch_s.lastepoch = ntohl(epoch_s.lastepoch);
		epoch_s.lasttime = time(NULL);

		switch(error){
			case 1: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_VERSION'");
				printf("PCP error : 'UNSUPP_VERSION'");
				#endif
				pcp_log("**'UNSUPP_VERSION' in get_cpe_mapping function\n");
				return 1;
			case 2: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'MALFORMED_REQUEST'");
				printf("PCP error : 'MALFORMED_REQUEST'");
				#endif
				pcp_log("**'MALFORMED_REQUEST' in get_cpe_mapping function\n");
				return 2;
			case 3: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_OPCODE'");
				printf("PCP error : 'UNSUPP_OPCODE'");
				#endif
				pcp_log("**'UNSUPP_OPCODE' in get_cpe_mapping function\n");
				return 3;
			case 4: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_OPTION'");
				printf("PCP error : 'UNSUPP_OPTION'");
				#endif
				pcp_log("**'UNSUPP_OPTION' in get_cpe_mapping function\n");
				return 4;
			case 5: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'MALFORMED_OPTION'");
				printf("PCP error : 'MALFORMED_OPTION'");
				#endif
				pcp_log("**'MALFORMED_OPTION' in get_cpe_mapping function\n");
				return 5;
			case 6: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'PROCESSING_ERROR'");
				printf("PCP error : 'PROCESSING_ERROR'");
				#endif
				pcp_log("**'PROCESSING_ERROR' in get_cpe_mapping function\n");
				return 6;
			case 7: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'SERVER_OVERLOADED'");
				printf("PCP error : 'SERVER_OVERLOADED'");				
				#endif
				pcp_log("**'SERVER_OVERLOADED' in get_cpe_mapping function\n");
				return 7;
			case 20:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NETWORK_FAILURE'");
				printf("PCP error : 'NETWORK_FAILURE'");
				#endif
				pcp_log("**'NETWORK_FAILURE' in get_cpe_mapping function\n");
				return 20;
			case 21: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NO_RESOURCES'");
				printf("PCP error : 'NO_RESOURCES'");
				#endif
				pcp_log("**'NO_RESOURCES' in get_cpe_mapping function\n");
				return 21;
			case 22: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_PROTOCOL'");
				printf("PCP error : 'UNSUPP_PROTOCOL'");
				#endif
				pcp_log("**'UNSUPP_PROTOCOL' in get_cpe_mapping function\n");
				return 22;
			case 23: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NOT_AUTHORIZED'");
				printf("PCP error : 'NOT_AUTHORIZED'");
				#endif
				pcp_log("**'NOT_AUTHORIZED' in get_cpe_mapping function\n");
				return 23;
			case 24: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'USER_EX_QUOTA'");
				printf("PCP error : 'USER_EX_QUOTA'");
				#endif
				pcp_log("**'USER_EX_QUOTA' in get_cpe_mapping function\n");
				return 24;
			case 25: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'CANNOT_PROVIDE_EXTERNAL_PORT'");
				printf("PCP error : 'CANNOT_PROVIDE_EXTERNAL_PORT'");
				#endif
				pcp_log("**'CANNOT_PROVIDE_EXTERNAL_PORT' in get_cpe_mapping function\n");
				return 25;
			case 26: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'EXCESSIVE_REMOTE_PEERS'");
				printf("PCP error : 'EXCESSIVE_REMOTE_PEERS'");
				#endif
				pcp_log("**'EXCESSIVE_REMOTE_PEERS' in get_cpe_mapping function\n");
				return 26;
			case 27:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'IMPLICIT_MAPPING_EXIT'");

				printf("PCP error : 'IMPLICIT_MAPPING_EXIT'\n");

				#endif
				pcp_log("**'IMPLICIT_MAPPING_EXIT' in get_cpe_mapping function\n");
				return 27;
			case 28:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'ADDRESS_MISMATCH'");
				printf("PCP error : 'ADDRESS_MISMATCH'");
				#endif
				pcp_log("**'ADDRESS_MISMATCH' in get_cpe_mapping function\n");
				return 28;
			case 130:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NO_MORE_MAPPING'");
				printf("PCP error : 'NO_MORE_MAPPING'");
				#endif
				pcp_log("**'NO_MORE_MAPPING' in get_cpe_mapping function\n");
				return 130;
			default: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "Undefined PCP errors ");
				printf("PCP error : 'Undefined PCP errors'");
				#endif
				pcp_log("**'Undefined PCP errors' in get_cpe_mapping function\n");
				return -1;					
			}
	}


	// 68bytes = 28bytes common-response-header + 40bytes GET opcode
	//get all next_option
	memcpy(next_option, rbuf +68, recv_len - 68);// 68bytes = 28bytes common-response-header + 40bytes GET opcode

	now = next_option;

	i =0;
	
	while(1){//separate next_options
		rcv_descr_len =0;
		des_len_in4 = 0;
		next_len =0;
		descr_flag = '\0';
		

		memcpy(&next_len, now+2, 2);
		next_len = ntohs(next_len);

		
		proto = *(now+4);
		in_af = *(now+5);
		ex_af = *(now+6);
		flag = *(now+7);


		if(in_af == 1){//ipv4
			/*internal IP address*/
			if (inet_ntop(AF_INET, now + 8, c_inip_4, sizeof(c_inip_4)) <= 0) {
				syslog(LOG_WARNING, "get_cpe_mapping(inet_ntop)");
				return 0;
			}
		}
		else if(in_af == 2){//ipv6
			if (inet_ntop(AF_INET6, now + 8, c_inip_6, sizeof(c_inip_6)) <= 0) {
				syslog(LOG_WARNING, "get_cpe_mapping(inet_ntop)");
				return 0;
			}

		}
		else{
			syslog(LOG_WARNING, "Error Address Family");
			return 0;
		}

		if(ex_af == 1){//ipv4
			/*external IPv4 address*/
			if (inet_ntop(AF_INET, now + 24, c_exip_4, sizeof(c_exip_4)) <= 0) {
				syslog(LOG_WARNING, "get_cpe_mapping(inet_ntop)");
				return 0;
			}

		}
		else if(ex_af == 2){//ipv6
			/*external IPv6 address*/
			if (inet_ntop(AF_INET6, now + 24, c_exip_6, sizeof(c_exip_6)) <= 0) {
				syslog(LOG_WARNING, "get_cpe_mapping(inet_ntop)");
				return 0;
			}

		}
		else{
			syslog(LOG_WARNING, "Error Address Family");
			return 0;
		}

		memcpy(&lifetime, now+40, 4);
		lifetime = ntohl(lifetime);

		s_inport = 0, s_export = 0;
		memcpy(&s_inport, now+44, 2);
		memcpy(&s_export, now+46, 2);
		s_inport = ntohs(s_inport);
		s_export = ntohs(s_export);	
		
		flag_len -= 48;
		next_len -= 48;

		descr_flag = *(now +48);
		
		if(descr_flag == DESCRIPTION_OPTION){
			printf("\t\t@@@DESCRIPTION APPEARS\n");
			if (next_len > 0){
				memcpy(&rcv_descr_len, now + 50, sizeof(rcv_descr_len));
				rcv_descr_len = ntohs(rcv_descr_len);

				//descr_buffer, dynamic fen pei descr_buffer nei cun
		
				if(rcv_descr_len > 0)
				{
					memcpy(descr_buffer, now + 52, rcv_descr_len);
				}
			}
			des_len_in4 = (rcv_descr_len+3)/4;
			des_len_in4 = des_len_in4*4;
			flag_len = flag_len - des_len_in4 -4;
	
		}
		else{
			printf("\t\t@@@NO DESCRIPTION!\n");
			memset(descr_buffer, 0, sizeof(descr_buffer));
			memcpy(descr_buffer, "NULL", 4);
		}
		if(in_af == 1 && ex_af ==1){//in&ex IPs are ipv4 addresses
			add_mapping_local(s_export, c_inip_4, c_exip_4, s_inport, proto, lifetime, descr_buffer);
		}
		else if(in_af == 1 && ex_af ==2){//in IP for ipv4 , ex IP for ipv6 not support yet

		}
		else if(in_af == 2 && ex_af ==1){//in IP for ipv6 , ex IP for ipv4 not support yet

		}
		else if(in_af == 2 && ex_af ==2){//in&ex IPs are ipv6 addresses not support yet

		}


		if(flag!=0){//flag 0 for END
			printf("$$$$$\t\nbreak\n");
			break;
		}
		if(flag_len <= 0){//need to resend attached last next_option
		
			printf("\t\n$$$$$resend\n");
			if(descr_flag == DESCRIPTION_OPTION){
				d_len = 4+des_len_in4;
			}
			else{
				d_len = 0;
			}
			memcpy(sbuf +68, now, 48+d_len);
			final_next_len = 48+d_len;
			goto re_send;
		}

		if(descr_flag == DESCRIPTION_OPTION){
			d_len = 4+des_len_in4;
		}
		else{
			d_len = 0;
		}

		now = now +48+d_len;
		
	}

	return 1;	
		
}

/* from netfilter/iptcrdr.c */

/*static struct rdr *
add_redirect_internal(unsigned short eport, const char *iaddr,
		      unsigned short iport, int proto,
		      uint32_t life, const char *desc)*/


static void
del_mapping_local(unsigned short ex_port, int protocol)
{
	struct rdr *p, *last;

	p = rdr_list;
	last = 0;
	
	#ifdef DEBUG_INFO
	printf("*****del_redirect_internal:\n");
	#endif
	syslog(LOG_WARNING, "*****del_redirect_internal:");
	
	while (p) {
		if (p->eport == ex_port && p->proto == protocol) {
			if (!last)
				rdr_list = p->next;
			else
				last->next = p->next;
			free(p);
			return;
		}
		last = p;
		p = p->next;
	}
}

static struct rdr *
get_mapping_local(unsigned short ex_port, int protocol)
{
	struct rdr *p;

	#ifdef DEBUG_INFO
	printf("*****get_redirect_internal:\n");
	#endif
	syslog(LOG_WARNING, "*****get_redirect_internal:");
	
	for (p = rdr_list; p; p = p->next) {
		if (p->eport == ex_port && p->proto == (short)protocol)
			return p;
	}
	
	#ifdef DEBUG_INFO
	printf("*****List_internal rules:\n");
	#endif

	
	for (p = rdr_list; p; p = p->next) {
		#ifdef DEBUG_INFO
		printf("*****List rules: eport==%d, proto==%d\n",p->eport, p->proto);
		#endif 
	}
	/**/
	return 0;
}




/* add rdr list to PCP server for Server rebuilding, by ZT */
int 
add_rdrlist_CGN (void)
{	
	unsigned char sbuf[128], rbuf[128]; 
	int error,cc;
	unsigned short len = 0;
	uint32_t epoch_new, tm;
	char pcp_cli_ip[INET6_ADDRSTRLEN];
	unsigned short s_inport = 0, s_export = 0;

	memset(sbuf, 0, sizeof(sbuf));


	struct rdr *p, *last, * p_copy;

	p = rdr_list;
	last = 0;
		
	#ifdef DEBUG_INFO
	printf("*****add_rdrlist_server:\n");
	#endif
	syslog(LOG_WARNING, "*****add_rdrlist_server:");
	
	while (p) {

		memset(sbuf, 0, sizeof(sbuf));

		sbuf[0] = 0x01; //version
		sbuf[1] = 0x01; //opcode
	
		tm = htonl(p->lifetime);
		memcpy(sbuf + 4, &tm, 4);

			//PCP client IP address
		if (inet_pton(AF_INET6, xnatpmp_local, pcp_cli_ip) <= 0) {
			syslog(LOG_ERR, "Failed to convert '%s' to local IPv6 address", xnatpmp_local);
		}

		memcpy(sbuf +12, pcp_cli_ip, 16);
		
		sbuf[28] =  p->proto == IPPROTO_TCP ? 6 : 17; //protocol
		sbuf[29] =0;
		sbuf[30] =0;
		sbuf[31] =0;
		
		s_inport = htons(p->iport);
		memcpy(sbuf +32, &s_inport, sizeof(s_inport));
		s_export = htons(p->eport);
		memcpy(sbuf +34, &s_export, sizeof(s_export));
		
		
		//external IP
		if (inet_pton(AF_INET, p->eaddr, sbuf + 36) <= 0) {
			syslog(LOG_WARNING, "add_redirect_rule2(inet_pton)");
			#ifdef DEBUG_INFO
			printf("Err:add_redirect_rule2 before sending (inet_pton)->eaddr\n");
			#endif
			return 0;
		}

		sbuf[40] = 0x03;//prefer-failure
	

		sbuf[44] = 0x04;//Third_Party option-code for 12th version(5th-version it's value is OX83)
		//option length, in units of 1 octets for 12th version( in units of 4 octets)
		unsigned short third_ip_len = 4;
		//unsigned short third_ip_len = 1;
		third_ip_len = htons(third_ip_len);
		memcpy(sbuf +46, &third_ip_len, sizeof(third_ip_len));
		//internal IP
		if (inet_pton(AF_INET, p->iaddr, sbuf + 48) <= 0) {
				syslog(LOG_WARNING, "add_redirect_rule2(inet_pton)");
				#ifdef DEBUG_INFO
				printf("Err:add_redirect_rule2 before sending (inet_pton)->iaddr\n");
				#endif
				return 0;
		}
		
		/* for description */
		if (NULL != p->desc)
		{
		   len = strlen(p->desc);
		   len = (len + 3)/4;
		   len = len *4;
		   if ((56 + len) <= 128)
		   {
			 sbuf[52]= 0x40;//desc-option; FIX maybe
			 int k = htons(len);
			 memcpy(sbuf +54, &k, sizeof(unsigned short));
			 memset(sbuf +56, 0, len);
			 memcpy(sbuf +56, p->desc, strlen(p->desc));
			 len+=4;
		   }
		   else
		   {
			   len = 0;
		   }
		}

		 if (send(sxfd6, sbuf, 52 + len, 0) < 0) {
			  syslog(LOG_WARNING, "add_redirect_rule2(send): %m");
			  #ifdef DEBUG_INFO
			  printf("Err:add_redirect_rule2(sending): %m\n");
			  #endif
			  return 0;
		 }

		memset(rbuf, 0, sizeof(rbuf));

		if ((cc=recv(sxfd6, rbuf, sizeof(rbuf), 0)) < 0) {
			syslog(LOG_WARNING, "add_redirect_rule2(recv): %m");
			#ifdef DEBUG_INFO
			printf("Err:add_redirect_rule2(recving): %m\n");
			#endif
			return 0;
		} 

		error = rbuf[3];
		
		memcpy(&epoch_new, rbuf + 8, 4);
		epoch_new = ntohl(epoch_new);

		epoch_s.lastepoch = epoch_new;
		epoch_s.lasttime = time(NULL);

		

		if (error != 0) {
			p_copy = p;

			#ifdef DEBUG_INFO
			printf("this mapping can't be rebulit.\n");	
			#endif
			if (!last)
				rdr_list = p->next;
			else
				last->next = p->next;
			p = p->next;

			free(p_copy);
			
			
		}
		else{
				memcpy(&p->lifetime, rbuf + 4, 4);
				p->lifetime = ntohl(p->lifetime);
				last = p;
				p = p->next;
		}

	
	}

	return 0;
}



	


struct rdr *
add_mapping_CGN(const char *ifname, unsigned short ex_port,
                   const char *in_ip, unsigned short in_port, int protocol,
		   unsigned int lifetime,
		   const char *desc, unsigned short *rcode)
{

	struct rdr *p;
	/*DXH*///unsigned char sbuf[20], rbuf[128];
	unsigned char sbuf[256], rbuf[1024];
	char error;
	unsigned short len = 0;
	uint32_t tm, lt, epoch_new, error_lf = 0;
	/*DXH*/
	char pcp_cli_ip[INET6_ADDRSTRLEN];
	char eaddr[INET_ADDRSTRLEN];

	char riaddr[INET_ADDRSTRLEN];


	unsigned short s_inport, s_export;

	#ifdef DEBUG_INFO
	printf("*****add_redirect_rule2:\n");
	syslog(LOG_WARNING, "*****add_redirect_rule2:");
	#endif

	memset(sbuf, 0, sizeof(sbuf));
	/*DXH*/
	//DXH start sending a request
	


	memset(sbuf, 0, sizeof(sbuf));
	memset(rbuf, 0, sizeof(rbuf));

	memset(pcp_cli_ip, 0, sizeof(pcp_cli_ip));

	sbuf[0] = 0x01; //version
	sbuf[1] = 0x01; //opcode request R bit =0 & MAP4 =1
	
	
	if (lifetime)
		tm = (uint32_t)lifetime;
	else
		tm = 3600U;
	tm = htonl(tm);
	memcpy(sbuf + 4, &tm, 4);

	//PCP client IP address
	if (inet_pton(AF_INET6, xnatpmp_local, pcp_cli_ip) <= 0) {
		syslog(LOG_ERR, "Failed to convert '%s' to local IPv6 address", xnatpmp_local);
		return NULL;
	}

	memcpy(sbuf +12, pcp_cli_ip, 16);
	


	sbuf[28] =  protocol == IPPROTO_TCP ? 6 : 17; //protocol
	sbuf[29] =0;
	sbuf[30] =0;
	sbuf[31] =0;

	s_inport= htons(in_port);
	memcpy(sbuf +32, &s_inport, sizeof(s_inport));
	s_export = htons(ex_port);
	memcpy(sbuf +34, &s_export, sizeof(s_export));

	//external IP
	if (inet_pton(AF_INET, "0.0.0.0", sbuf + 36) <= 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "add_redirect_rule2(inet_pton)");
		printf("Err:add_redirect_rule2 before sending (inet_pton)->eaddr\n");
		#endif
		return NULL;
	}

	
	sbuf[40] = 0x03;//prefer-failure
	

	sbuf[44] = 0x04;//Third_Party option-code for 12th version(5th-version it's value is OX83)
	//option length, in units of 1 octets for 12th version( in units of 4 octets for 5th version)
	//unsigned short third_ip_len = 1;
	unsigned short third_ip_len = 4;
	third_ip_len = htons(third_ip_len);
	memcpy(sbuf +46, &third_ip_len, sizeof(third_ip_len));


	//internal IP
	#ifdef MY_DEBUG
	printf("\t\t@@@@@@@add internal ip address :%s\n", in_ip);
	#endif
	if (inet_pton(AF_INET, in_ip, sbuf + 48) <= 0) {
			#ifdef DEBUG_INFO
			syslog(LOG_WARNING, "add_redirect_rule2(inet_pton)");
			printf("Err:add_redirect_rule2 before sending (inet_pton)->iaddr\n");
			#endif
			return NULL;
	}

/* for description */

   if (NULL != desc)
   {
      len = strlen(desc);
	  len = (len + 3)/4;
	  len = len *4;
      if ((56 + len) <= 128)
      {
      	sbuf[52]= 0x40;//desc-option; FIX maybe
		int k = htons(len);
	    memcpy(sbuf +54, &k, sizeof(unsigned short));
		memset(sbuf +56, 0, len);
		memcpy(sbuf +56, desc, strlen(desc));
		len+=4;
      }
      else
      {
          len = 0;
      }
   }

	
	if (send(sxfd6, sbuf, 52 + len, 0) < 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "add_redirect_rule2(send): %m");
		printf("Err:add_redirect_rule2(sending): %m\n");
		#endif
		return 0;
	}

	#ifdef MY_DEBUG
	printf("\n\t#########add redirect send %d:\n",(52+len));
	print_packet_content(sbuf, (52+len));
	#endif

		/*TODO: PCP Request Address Family/PCP Request Port/PCP Request IP Address fileds  */

	memset(rbuf, 0, sizeof(rbuf));

	int cc;
	
	#ifdef MY_DEBUG
	printf("\n\t#########start to add redirect recv \n");
	#endif
	if ((cc=recv(sxfd6, rbuf, 1024, 0))<0)
	{
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "add_redirect_rule2(recv): %m");
		printf("Err:add_redirect_rule2(recieve): %m\n");
		#endif
		return 0;
	}

	#ifdef MY_DEBUG
	printf("\n\t#########add redirect recv:cc\n");
	print_packet_content(rbuf, cc);
	#endif
	
	/*Life time*/
	memcpy(&lt, rbuf + 4, 4);
	lt = ntohl(lt);
	
	error = rbuf[3];
	
	#ifdef MY_DEBUG
	printf("\n\t#############result code :%d\n", error);
	#endif
	
	memcpy(&error_lf, rbuf +4, 4);
	error_lf = ntohl(error_lf);
	if (error != 0) {//todo: handle result code
		struct rdr *tmp;
		tmp = malloc(sizeof(struct rdr) + 1);
			
		if(tmp){
			memset(tmp, 0, sizeof(struct rdr)+1);
			tmp->flag = 0;
		}
		switch(error){
			case 1: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_VERSION'");
				printf("PCP error : 'UNSUPP_VERSION'\n");
				#endif
				pcp_log("**'UNSUPP_VERSION' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 2: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'MALFORMED_REQUEST'");
				
				printf("PCP error : 'MALFORMED_REQUEST'\n");
				#endif
				pcp_log("**'MALFORMED_REQUEST' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 3: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_OPCODE'");
				
				printf("PCP error : 'UNSUPP_OPCODE'\n");
				#endif
				pcp_log("**'UNSUPP_OPCODE' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 4: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_OPTION'");
				
				printf("PCP error : 'UNSUPP_OPTION'\n");
				#endif
				pcp_log("**'UNSUPP_OPTION' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 5: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'MALFORMED_OPTION'");
				
				printf("PCP error : 'MALFORMED_OPTION'\n");
				#endif
				pcp_log("**'MALFORMED_OPTION' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 6: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'PROCESSING_ERROR'");
				
				printf("PCP error : 'PROCESSING_ERROR'\n");
				#endif
				pcp_log("**'PROCESSING_ERROR' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 7: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'SERVER_OVERLOADED'");
				
				printf("PCP error : 'SERVER_OVERLOADED'\n");
				#endif
				pcp_log("**'SERVER_OVERLOADED' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				sleep(1800);
				return 0;
			case 20: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NETWORK_FAILURE'");
				
				printf("PCP error : 'NETWORK_FAILURE'\n");
				#endif
				pcp_log("**'NETWORK_FAILURE' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				tmp->flag = 1;
				tmp->sleep_tm =30;
				return tmp;
			case 21: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NO_RESOURCES'");
				printf("PCP error : 'NO_RESOURCES'\n");
				#endif
				pcp_log("**'NO_RESOURCES' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				tmp->flag = 1;
				tmp->sleep_tm =30;
				return tmp;
			case 22: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_PROTOCOL'");
				printf("PCP error : 'UNSUPP_PROTOCOL'\n");
				#endif
				pcp_log("**'UNSUPP_PROTOCOL' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 23: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NOT_AUTHORIZED'");
				printf("PCP error : 'NOT_AUTHORIZED'\n");
				#endif
				pcp_log("**'NOT_AUTHORIZED' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 24: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'USER_EX_QUOTA'");
				printf("PCP error : 'USER_EX_QUOTA'\n");
				#endif
				pcp_log("**'USER_EX_QUOTA' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				tmp->flag = 1;
				tmp->sleep_tm = 30;
				return tmp;
			case 25: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'CANNOT_PROVIDE_EXTERNAL_PORT' in %d", error_lf);
				printf("PCP error : 'CANNOT_PROVIDE_EXTERNAL_PORT'\n");
				#endif
				pcp_log("**'CANNOT_PROVIDE_EXTERNAL_PORT' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				printf("error_lf : %d\n", error_lf);
				if((error_lf>0)&&(error_lf<120)){
					tmp->flag = 1;
					tmp->sleep_tm = error_lf;
					return tmp;
				}
				else{
					return 0;
				}
			case 26: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'EXCESSIVE_REMOTE_PEERS'");

				printf("PCP error : 'EXCESSIVE_REMOTE_PEERS'\n");
				#endif
				pcp_log("**'EXCESSIVE_REMOTE_PEERS' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 27: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'IMPLICIT_MAPPING_EXIT'");

				printf("PCP error : 'IMPLICIT_MAPPING_EXIT'\n");
				#endif
				pcp_log("**'IMPLICIT_MAPPING_EXIT' when Add Mappings\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 28:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'ADDRESS_MISMATCH'");
				printf("PCP error : 'ADDRESS_MISMATCH'");
				#endif
				pcp_log("**'ADDRESS_MISMATCH' in get_cpe_mapping function\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			case 130:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NO_MORE_MAPPING'");
				printf("PCP error : 'NO_MORE_MAPPING'");
				#endif
				pcp_log("**'NO_MORE_MAPPING' in get_cpe_mapping function\n");
				if (rcode){
					*rcode = error;
				}
				return 0;
			default:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "Undefined PCP error ");
				printf("PCP error : 'Undefined PCP error'\n");
				#endif
				pcp_log("**'Undefined PCP error' in get_cpe_mapping function\n");
				if (rcode){
					*rcode = error;
				}
				return 0;					
			}
			
		}
		printf("Recived Process: Result Code OK\n");


	/* for server rebuilding, by ZT */
	memcpy(&epoch_new, rbuf + 8, 4);
	epoch_new = ntohl(epoch_new);

	
	#ifdef DEBUG_INFO
	printf( "last epoch: %d\n",epoch_s.lastepoch);
	printf( "last time: %d\n",epoch_s.lasttime);
	printf( "new epoch: %d\n",epoch_new);
	printf( "now time: %d\n",(uint32_t)time(NULL));
	printf( "time length: %d\n",(uint32_t)(time(NULL)-epoch_s.lasttime));
	#endif
	if (epoch_new < epoch_s.lastepoch+(time(NULL)-epoch_s.lasttime)*7/8)
		add_rdrlist_CGN ();
	else
	{
		epoch_s.lastepoch = epoch_new;
		epoch_s.lasttime = time(NULL);
	}


		/*Assigned external IP address*/
		if (inet_ntop(AF_INET, rbuf + 36, eaddr, sizeof(eaddr)) <= 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "add_redirect_rule2(inet_ntop)");
		printf("Err:add_redirect_rule2 external IP(inet_ntop):after send\n");
		#endif
		return 0;
		}
		
		#ifdef DEBUG_INFO
		printf("Assigned external IP address OK\n");
		#endif


		s_inport = 0;
		s_export = 0;
		memcpy(&s_inport, rbuf+32, 2);
		memcpy(&s_export, rbuf+34, 2);
		s_inport = ntohs(s_inport);
		s_export = ntohs(s_export);


		/*iport*/
		#ifdef MY_DEBUG
		printf("\t\t@@@@@@@@@@s_inport :%u  iport : %u\n", s_inport, in_port);
		#endif
		if (s_inport != in_port) {
			#ifdef DEBUG_INFO
			syslog(LOG_WARNING, "add_redirect_rule2: iport mismatch");
			printf("Err:add_redirect_rule2(returnd port '%d' doesn't match iport '%d')\n",(rbuf[40] << 8) | rbuf[41],  in_port);
			#endif
			return NULL;
		}

		/*eport*/
		/*what if eport mismatch for NAT-PMP and UpnP*/ /*UPnP 1.0 */
		if (rcode) {
			ex_port = s_export;
		} 
		else if (s_export != ex_port) {

			#ifdef DEBUG_INFO
			syslog(LOG_WARNING, "add_redirect_rule2: eport mismatch");
			
			printf("Err:add_redirect_rule2(returnd eport '%d' doesn't match iport '%d')\n",(rbuf[42] << 8) | rbuf[43],  ex_port);
			#endif
			return 0;
		}
		#ifdef DEBUG_INFO
		printf("eport OK\n");
		#endif
	
   if(0x04 == rbuf[44])
   {
   		/*Assigned internal IP address*/
		if (inet_ntop(AF_INET, rbuf + 48, riaddr, sizeof(riaddr)) <= 0) 
		{
			#ifdef DEBUG_INFO
			syslog(LOG_WARNING, "add_redirect_rule2(inet_ntop)");
			printf("Err:add_redirect_rule2 external IP(inet_ntop):after send\n");
			#endif
			return 0;
		}

		if (memcmp(riaddr, in_ip, 4) != 0)
		{
			#ifdef DEBUG_INFO
			printf("add_redirect_rule2: request internal IP != respond internal IP\n There is a NAT between?\n");
			#endif
			return 0;
		}

		#ifdef DEBUG_INFO
		printf("internal IP address from THIRD_PARTY option OK\n");
		#endif
		
   }
   else
   {
   		#ifdef DEBUG_INFO
  		printf("add_redirect_rule2: no responsed internal IP from server\n");
		#endif
		return 0;
   		
   }
   #ifdef MY_DEBUG
   printf("\n\t######add internal eport:%u, proto: %d\n", ex_port, protocol);
   #endif
   
   if(!(p = get_mapping_local(ex_port, protocol))){
   		return add_mapping_local(ex_port, in_ip, eaddr, in_port, protocol, lt, desc);
   }else{
		return NULL;
   }
}

struct rdr *
get_mapping(const char *ifname, unsigned short ex_port, int protocol)
{

	return get_mapping_local(ex_port, protocol);
}	

struct rdr *
get_mapping_by_index(int index, char *ifname)
{
	struct rdr *p = 0;
	int i = 0;

	/*DXH*/


	for (p = rdr_list; p; p = p->next) {

		if (i == index)
			break;
		i++;
		}
	return p;

	/*DXH*/

}




int
delete_mapping_CGN(unsigned short ex_port, int protocol)
{

		/*DXH*/
		
	struct rdr *p;
	unsigned char sbuf[52], rbuf[52];
	uint32_t epoch_new, error_lf =0;
	char pcp_cli_ip[INET6_ADDRSTRLEN];
	unsigned short s_inport = 0, s_export = 0;
	int error =0;
	
	#ifdef DEBUG_INFO
	syslog(LOG_WARNING, "*****delete_redirect_rule:");

	printf("\n\t#######delete protocol\n");
	#endif
	p = get_mapping_local(ex_port, protocol);
	
	if (!p)
	{
		#ifdef DEBUG_INFO
		printf("*****no_rules for : eport:%d, proto:%d \n", ex_port, protocol);
		#endif
		return 0;
	
	}
	memset(sbuf, 0, sizeof(sbuf));
	memset(rbuf, 0, sizeof(rbuf));
	
	//DXH start sending a request

	#ifdef DEBUG_INFO
	printf("*****delete_redirect_rule: eport:%d, proto:%d \n", ex_port, protocol);
	#endif


	/*CREATE REQ for delete*/
	sbuf[0] = 0x01; //version
	sbuf[1] = 0x01; //opcode
	

	//PCP client IP address
	if (inet_pton(AF_INET6, xnatpmp_local, pcp_cli_ip) <= 0) {
		syslog(LOG_ERR, "Failed to convert '%s' to local IPv6 address", xnatpmp_local);
		return 0;
	}

	memcpy(sbuf +12, pcp_cli_ip, 16);

	sbuf[28] =  protocol == IPPROTO_TCP ? 6 : 17; //protocol

	s_inport = htons(p->iport);
	memcpy(sbuf +32, &s_inport, sizeof(s_inport));
	s_export = htons(ex_port);
	memcpy(sbuf +34, &s_export, sizeof(s_export));

	//sbuf[40] = 0x03;//prefer_failure
	


 	sbuf[40] = 0x04;//third party
 	unsigned short third_ip_len = 4;//VERSION 12 LENGTH
 	//unsigned short third_ip_len = 1;
	third_ip_len = htons(third_ip_len);
	memcpy(sbuf +42, &third_ip_len, sizeof(third_ip_len));

	//internal IP
	if (inet_pton(AF_INET, p->iaddr, sbuf + 44) <= 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "delete_redirect_rule2(inet_pton)");
		
		printf("delete_redirect_rule: (inet_pton->iaddr)");
		#endif
		return 0;
	}

	#ifdef MY_DEBUG
   printf("\n\t#######delete redirect rule send:\n");
   print_packet_content(sbuf, 48);
   #endif
   
	if (send(sxfd6, sbuf, 48, 0) < 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "delete_redirect_rule(send): %m");
		
		printf( "delete_redirect_rule(send): %m");
		#endif
		return 0;
	}
	//DXH end sending a request


	memset(rbuf, 0, sizeof(rbuf));

	if (recv(sxfd6, rbuf, 48, 0)<0)
	{
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "del_redirect_rule(recv): %m");
		
		printf("Err:del_redirect_rule(recieve): %m\n");
		#endif
		return 0;
	}

	#ifdef MY_DEBUG
	printf("\n\t#######delete redirect rule recv:\n");
	print_packet_content(sbuf, 48);
	#endif

	error = rbuf[3];

	memcpy(&error_lf, rbuf +4, 4);
	error_lf = ntohl(error_lf);
	#ifdef MY_DEBUG
	printf("$$$$$$$error lifetime: %d\n", error_lf);
	#endif
	if (error != 0) {//todo: handle result code
		switch(error){
			case 1: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_VERSION'");
				
				printf("PCP error : 'UNSUPP_VERSION'");
				#endif
				pcp_log("**'UNSUPP_VERSION' when delete mapping\n");
				return 0;
			case 2: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'MALFORMED_REQUEST'");
				printf("PCP error : 'MALFORMED_REQUEST'");
				#endif
				pcp_log("**'MALFORMED_REQUEST' when delete mapping\n");
				return 0;
			case 3: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_OPCODE'");
				printf("PCP error : 'UNSUPP_OPCODE'");
				#endif
				pcp_log("**'UNSUPP_OPCODE' when delete mapping\n");
				return 0;
			case 4: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_OPTION'");
				printf("PCP error : 'UNSUPP_OPTION'");
				#endif
				pcp_log("**'UNSUPP_OPTION' when delete mapping\n");
				return 0;
			case 5: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'MALFORMED_OPTION'");
				printf("PCP error : 'MALFORMED_OPTION'");
				#endif
				pcp_log("**'MALFORMED_OPTION' when delete mapping\n");
				return 0;
			case 6: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'PROCESSING_ERROR'");
				printf("PCP error : 'PROCESSING_ERROR'");
				#endif
				pcp_log("**'PROCESSING_ERROR' when delete mapping\n");
				return 0;
			case 7: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'SERVER_OVERLOADED'");
				printf("PCP error : 'SERVER_OVERLOADED'");
				#endif
				pcp_log("**'SERVER_OVERLOADED' when delete mapping\n");
				sleep(1800);
				return 0;
			case 20:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NETWORK_FAILURE'");
				printf("PCP error : 'NETWORK_FAILURE'");
				#endif
				pcp_log("**'NETWORK_FAILURE' when delete mapping\n");
				return 0;
			case 21: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NO_RESOURCES'");
				printf("PCP error : 'NO_RESOURCES'");
				#endif
				pcp_log("**'NO_RESOURCES' when delete mapping\n");
				return 30;
			case 22: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'UNSUPP_PROTOCOL'");
				printf("PCP error : 'UNSUPP_PROTOCOL'");
				#endif
				pcp_log("**'UNSUPP_PROTOCOL' when delete mapping\n");
				return 0;
			case 23: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NOT_AUTHORIZED'");
				printf("PCP error : 'NOT_AUTHORIZED'");
				#endif
				pcp_log("**'NOT_AUTHORIZED' when delete mapping\n");
				return 0;
			case 24: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'USER_EX_QUOTA'");
				printf("PCP error : 'USER_EX_QUOTA'");
				#endif
				pcp_log("**'USER_EX_QUOTA' when delete mapping\n");
				return 30;
			case 25: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'CANNOT_PROVIDE_EXTERNAL_PORT' in %d seconds", error_lf);
				printf("PCP error : 'CANNOT_PROVIDE_EXTERNAL_PORT' in %d seconds", error_lf);
				#endif
				pcp_log("**'CANNOT_PROVIDE_EXTERNAL_PORT' when delete mapping\n");
				return error_lf;
			case 26: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'EXCESSIVE_REMOTE_PEERS'");
				printf("PCP error : 'EXCESSIVE_REMOTE_PEERS'");
				#endif
				pcp_log("**'EXCESSIVE_REMOTE_PEERS' when delete mapping\n");
				return 0;
			case 27:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'IMPLICIT_MAPPING_EXIT'");

				printf("PCP error : 'IMPLICIT_MAPPING_EXIT'\n");

				#endif
				pcp_log("**'IMPLICIT_MAPPING_EXIT' when delete mapping\n");
				return 0;
			case 28:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'ADDRESS_MISMATCH'");
				printf("PCP error : 'ADDRESS_MISMATCH'");
				#endif
				pcp_log("**'ADDRESS_MISMATCH' when delete mapping\n");
				return 0;
			case 130:
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "PCP error : 'NO_MORE_MAPPING'");
				printf("PCP error : 'NO_MORE_MAPPING'");
				#endif
				pcp_log("**'NO_MORE_MAPPING' when delete mapping\n");
				return 0;
			default: 
				#ifdef DEBUG_INFO
				syslog(LOG_WARNING, "Undefined PCP errors ");
				printf("PCP error : 'Undefined PCP errors'");
				#endif
				pcp_log("**'Undefined PCP errors' when delete mapping\n");
				return 0;					
			}
			
		}

	
	#ifdef DEBUG_INFO
	printf( "received a response for deleting a mapping.");
	#endif
	del_mapping_local(ex_port, protocol);

	/* for server rebuilding, by ZT */
	memcpy(&epoch_new, rbuf + 8, 4);
	epoch_new = ntohl(epoch_new);

	#ifdef DEBUG_INFO
	printf( "last epoch: %d\n",epoch_s.lastepoch);
	printf( "last time: %d\n",epoch_s.lasttime);
	printf( "new epoch: %d\n",epoch_new);
	printf( "now time: %d\n",(uint32_t)time(NULL));
	printf( "time length: %d\n",(uint32_t)(time(NULL)-epoch_s.lasttime));
	#endif
	if (epoch_new < epoch_s.lastepoch+(time(NULL)-epoch_s.lasttime)*7/8)
		add_rdrlist_CGN ();
	else
	{
		epoch_s.lastepoch = epoch_new;
		epoch_s.lasttime = time(NULL);
	}

	
	return 0;


}



int
delete_mappings(const char *in_ip, int protocol)
{
	int r = 0;
	int i=0;
	unsigned char buf[44];
	struct rdr *p, *last;


	memset(buf, 0, sizeof(buf));
	

	buf[0] = 0x01; //version
	buf[1] = 0x01; //opcode
	buf[2] =  protocol == IPPROTO_TCP ? 2 : 1; //protocol
	buf[3] = 0x0; //reserved 
	for (i=4;i<28;i++) //reserverd
		buf[i] = 0;
	
	if (send(sxfd6, buf, 44, 0) < 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "delete_redirect_rule(send): %m");
		printf( "delete_redirect_rule(send): %m");
		#endif
		r = -1;
	} 
	

	p = rdr_list;
	last = 0;
	while (p) {
		if (strcmp(p->iaddr, in_ip) == 0 && p->proto == protocol) {
			struct rdr *next = p->next;
			if (!last)
				rdr_list = next;
			else
				last->next = next;
			free(p);
			p = next;
		} else {
			last = p;
			p = p->next;
		}
	}
	return r;
}


int
del_all_mappings_CGN(void)//send a request to delete all mappings on PCP-server when shutdown PCP-client
{
	unsigned char sbuf[48], rbuf[48];
	int error = 0;

	char pcp_cli_ip[INET6_ADDRSTRLEN];
	
	

	memset(sbuf, 0, sizeof(sbuf));

	#ifdef DEBUG_INFO
	syslog(LOG_WARNING, "*****del_all_rule:");
	printf("*****del_all_rule error \n");
	#endif

	/*CREATE REQ for delete all ,lifetime=0, protocol=0,IN&EX-port=0,ex-ip=0,third-party in-ip =0*/
	sbuf[0] = 0x01; //version
	sbuf[1] = 0x01; //opcode
	

	//PCP client IP address
	if (inet_pton(AF_INET6, xnatpmp_local, pcp_cli_ip) <= 0) {
		syslog(LOG_ERR, "Failed to convert '%s' to local IPv6 address", xnatpmp_local);
		return 0;
	}

	memcpy(sbuf +12, pcp_cli_ip, 16);

 	sbuf[40] = 0x04;//third party
 	unsigned short third_ip_len = 4;
	third_ip_len = htons(third_ip_len);
	memcpy(sbuf +42, &third_ip_len, sizeof(third_ip_len));
	
	printf("***send del_all\n");
	if (send(sxfd6, sbuf, sizeof(sbuf), 0) < 0) {
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "delete_redirect_rule(send): %m");
		
		printf( "del_all_rule(send): %m");
		#endif
		return 0;
	}

	memset(rbuf, 0, sizeof(rbuf));

	#ifdef MY_DEBUG
	printf("***wait for recv del_all\n");
	#endif
	
	if (recv(sxfd6, rbuf, sizeof(rbuf), 0)<0)
	{
		#ifdef DEBUG_INFO
		syslog(LOG_WARNING, "del_all_rule(recv): %m");
		
		printf("Err:del_all_rule(recieve): %m\n");
		#endif
		return 0;
	}

	#ifdef DEBUG_INFO

	printf( "\nreceived a response for del_all mappings.\n");
	print_packet_content(rbuf, sizeof(rbuf));

	#endif

	error = rbuf[3];

	if(error == 0){
		#ifdef DEBUG_INFO
		printf( "\nend del_all mappings.\n");
		#endif
		return 0;
	}
	else{
		#ifdef DEBUG_INFO
		printf("result code error!\n");
		#endif
		return -1;
	}

}





/* for PCP refresh by ZT */
int
refresh_mappings(struct rdr *r)
{
	unsigned char sbuf[128], rbuf[128];
	int cc;
	int n=0;
	unsigned short len = 0;
	uint32_t tm, lt;
	
	char pcp_cli_ip[INET6_ADDRSTRLEN];
	unsigned short s_inport =0, s_export = 0;

	#ifdef DEBUG_INFO
	printf("*****PCP refresh:\n");
	
	syslog(LOG_WARNING, "*****CP refresh:");
	#endif



	memset(sbuf, 0, sizeof(sbuf));
	sbuf[0] = 0x01; //version
	sbuf[1] = 0x01; //opcode

	tm = htonl(r->lifetime);
	memcpy(sbuf + 4, &tm, 4);


		//PCP client IP address
	if (inet_pton(AF_INET6, xnatpmp_local, pcp_cli_ip) <= 0) {
		syslog(LOG_ERR, "Failed to convert '%s' to local IPv6 address", xnatpmp_local);
	}

	memcpy(sbuf +12, pcp_cli_ip, 16);

	sbuf[28] = r->proto == IPPROTO_TCP ? 6 : 17;

	s_inport = htons(r->iport);
	memcpy(sbuf +32, &s_inport, sizeof(s_inport));
	s_export = htons(r->eport);
	memcpy(sbuf +34, &s_export, sizeof(s_export));



	if (inet_pton(AF_INET, r->eaddr, sbuf + 36) <= 0) {
		syslog(LOG_WARNING, "refresh_redirect(inet_pton)");
		return -3;
	}


	sbuf[40] = 0x03;//prefer-failure
	
	
	sbuf[44] = 0x04;//Third_Party option-code for 12th version(5th-version it's value is OX83)
	//option length, in units of 1 octets for 12th version( in units of 4 octets)
	unsigned short third_ip_len = 4;
	//unsigned short third_ip_len = 1;
	third_ip_len = htons(third_ip_len);
	memcpy(sbuf +46, &third_ip_len, sizeof(third_ip_len));
	//internal IP
	if (inet_pton(AF_INET, r->iaddr, sbuf + 48) <= 0) {
			#ifdef DEBUG_INFO
			syslog(LOG_WARNING, "add_redirect_rule2(inet_pton)");
			
			printf("Err:add_redirect_rule2 before sending (inet_pton)->iaddr\n");
			#endif
			return -3;
	}


	
	/* for description */
	if (NULL != r->desc)
	{
	   len = strlen(r->desc);
	   len = (len + 3)/4;
	   len = len * 4;
	   if ((56 + len) <= 128)
	   {
		 sbuf[52]= 0x40;//desc-option; FIX maybe
		 int k = htons(len);
		 memcpy(sbuf +54, &k, sizeof(unsigned short));
		 memset(sbuf +56, 0, len);
		 memcpy(sbuf +56, r->desc, strlen(r->desc));
		 len+=4;
	   }
	   else
	   {
		   len = 0;
	   }
	}


		
	if (send(sxfd6, sbuf, 52 + len, 0) < 0) {

		syslog(LOG_WARNING, "PCP refresh(send): %m");
		return -3;
	}

	/*Re-send*/
	fd_set readset;	
	int max_fd = -1;
	struct timeval timeout;
	int ret = 0;
	
	if (sxfd6 >= 0) {
		FD_ZERO(&readset);
		FD_SET(sxfd6, &readset);
		max_fd = (max_fd > sxfd6)? max_fd:sxfd6;
		timeout.tv_sec = 5;
		timeout.tv_usec = 0;

		wait_response:
		ret = select(max_fd+1, &readset, 0, 0, &timeout);
		
		if (ret < 0) {
				 perror("select");
				 return 0;
			 } else if (ret == 0) {
				 printf("timeout\n");
				 return 0;
			 }else
			 {
				 memset(rbuf, 0, sizeof(rbuf));
				 cc = recv(sxfd6, rbuf, 128, MSG_PEEK);

					printf("\n\t#########refresh redirect recv:\n");
					print_packet_content(sbuf, cc);
				 

				 if (cc < 0) 
				 	{
				 		#ifdef DEBUG_INFO
				    	 syslog(LOG_WARNING, "PCP refresh(recv): %m");
						 
						 printf("PCP refresh (recieve): %m\n");
						 #endif
						 return 0;
					 }
				sbuf[1]+=0x80;

				 if ((cc != (52+len)) || (memcmp(sbuf, rbuf, 3) != 0))
				{

					#ifdef DEBUG_INFO
					printf("**PCP refresh resopnse version and opcode doesn't match\n");
					printf("recieved byte == %d\n",cc);
					printf("sbuf [0]: %d\n",sbuf[0]);
					printf("sbuf [1]: %d\n",sbuf[1]);
					printf("sbuf [2]: %d\n",sbuf[2]);

					printf("rbuf [0]: %d\n",rbuf[0]);
					printf("rbuf [1]: %d\n",rbuf[1]);
					printf("rbuf [2]: %d\n",rbuf[2]);
					#endif
					/* resend */
					sbuf[1] = 0x01;
					if (send(sxfd6, sbuf, 52 + len, 0) < 0) {
						#ifdef DEBUG_INFO
						syslog(LOG_WARNING, "PCP refresh resopnse(resend): %m");
						
						printf("Err:PCP refresh resopnse(resending): %m\n");
						#endif
						return 0;
					}
					else{ 
						timeout.tv_sec = 2* timeout.tv_sec;
						n=n+1;
					}

					if (n <2) goto wait_response;
					else return 0;


				}

				

			 }
	}

	memcpy(&lt, rbuf + 4, 4);
	lt = ntohl(lt);
	r->lifetime = lt;
	tm = (uint32_t)(time(NULL) - startup_time);
	if (r->expire)
		r->expire = tm + lt;


	if (r->refresh)
		r->refresh = tm + lt / 2;  
		
	return 0;
}



uint32_t
manage_mappings(uint32_t lifetime)
{
	struct rdr *r;
	int i;


	/* expire loop */
	i = 0;
	for(;;) {
		r = get_mapping_by_index(i, 0);
		if (r == 0)
			break;
		if (r->expire != 0 && r->expire < lifetime)
			del_mapping_local(r->eport, r->proto);
		else
			i++;
	}

	/* refresh loop */
	i = 0;
	for (;;) {
		r = get_mapping_by_index(i, 0);
		if (r == 0)
			break;
		if (r->refresh != 0 && r->refresh < lifetime)
			(void) refresh_mappings(r);
		i++;
	}

	/* next refresh */
	lifetime = 0;
	i = 0;
	for (;;) {
		r = get_mapping_by_index(i, 0);
		if (r == 0)
			break;
		if (r->refresh != 0 && (lifetime == 0 || r->refresh < lifetime))
			lifetime = r->refresh;
		i++;
	}
	return lifetime;
}
