/* 
* Module Name : upnp-pcp iwf
 
* Version : 01
 
*
 
* Software Name : upnp-pcp interworking function according to pcp base draft -12
 
* Version : 01
 
*
 
 
 
* Copyright: 10/06/2011  France Telecom
 
* This software is distributed under the Apache 2.0 license,
 
* the text of which is available at http://www.apache.org/licenses/LICENSE-2.0.html
 
 
 
*
 
*--------------------------------------------------------
 
* File Name   : pcp.h
 
*
 
* Created     : 06/2011
 
* Author(s)   : Xiaohong Deng, Tao Zheng, Zhengyu Lei
 
*
 
* Description :
 
* pcp_log : record pcp error information
* add_mapping_local : add mapping (five tuples, lifetime and descriptions)  to memory
* get_cpe_mapping : get all mappings from CGN
* del_mapping_local : delete mapping from memory
* get_mapping_local : get a specific mapping from memory
* get_mapping : get a specific mapping from CGN
* add_rdrlist_CGN : add local mappings to CGN for Server rebuilding
* add_mapping_CGN : send a specific pcp add-mapping request to CGN
* get_mapping_by_index : get a specific mapping by index
* delete_mapping_CGN : send a specific pcp delete-mapping request to CGN
* delete_mappings : send request to delete mappings
* del_all_mappings_CGN : send request to delete all mappings from this CPE in CGN
* refresh_mappings : refresh mappings 
*--------------------------------------------------------
*/

#include "config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

void 
print_packet_content(unsigned char *pBuf, int bufLen);

int
pcp_log(const char *string);


int 
get_cpe_mapping(void);


int 
add_rdrlist_CGN (void);


struct rdr *
add_mapping_CGN(const char *ifname, unsigned short ex_port,
                   const char *in_ip, unsigned short in_port, int protocol,
		   unsigned int lifetime, const char *desc, unsigned short *rcode);


struct rdr *
get_mapping(const char *ifname, unsigned short ex_port, int protocol);

struct rdr *
get_mapping_by_index(int index, char *ifname);

int
delete_mapping_CGN(unsigned short ex_port, int protocol);

int
delete_mappings(const char *in_ip, int protocol);

int
del_all_mappings_CGN(void);

int
refresh_mappings(struct rdr *r);

uint32_t
manage_mappings(uint32_t lifetime);

